module Unix = UnixLabels
module Stringmap = Map.Make (String)
module Stringset = Set.Make (String)
open Printf
open Str

open Mis
open Logger
open Clientinfo

module Connection =
  struct

		type tPacketType = HeaderOnly | Chunked | ContentLength | ContentType | Eof | PacketTypeError;;

		type tRequestHeaderType = Get | Head | Post | Put | Delete | Trace | RHTError;;

		type role = Client | Server;;

		exception NoData

    (* uline is the url line, including the "get" keyword  *)
    class urlLine =
    object
      val vUrlline = new Mis.misString;
      val keyword = new Mis.misString;
      val vUrl = new Mis.misString;
      val mutable port = 0;
      val portregex = (regexp ":\\([^/]+\\)/");

			method seturl uline = vUrlline#setstr uline;

			method process =
				try
					begin
						let fs = Str.split (Mis.whitespace()) vUrlline#asString in
							keyword#setstr (List.nth fs 0);
							vUrl#setstr (List.nth fs 1);
							try
								Str.search_forward portregex (vUrl#asString) 0;
								port <- int_of_string (matched_string vUrl#asString);
							with Not_found ->
								ignore();
					end
				with x ->
					Logger.showex "connection.ml" 46 x "exception in process";
    end;;

		type tTransfer = Request | Response;;

		class httpDataTransfer transactionType =
			object(self)
				inherit urlLine as super
				val mutable hmap = Stringmap.empty;

				method lookup str = Stringmap.find str hmap

				method headerFromStringInternal (lines:string list) =
					Logger.show 5 "connection.ml" 59 ("headerFromStringInternal, lines=" ^ (Mis.stringOfStringList lines ""));

					try
						begin
							for loop = 1 to (List.length lines) do
								let li = (List.nth lines (loop-1)) in
								let lilc = String.lowercase li in
									Logger.show 3 "connection.ml" 66 ("lilc=" ^ lilc);
									if (String.length lilc) > 0
									then
										let dli = String.lowercase lilc in
										let fs = Str.bounded_split (regexp "[ \t]+") dli 2 in
											hmap <- (Stringmap.add (List.nth fs 0) (List.nth fs 1) hmap)
							done
						end
					with x ->
						Logger.showex "connection.ml" 75 x "exception in headerFromStringInternal";

				method header_from_string lines =
					Logger.show
						5
						"connection.ml"
						65
						("header_from_string, lines=" ^ (Mis.stringOfStringList lines ""));

					self#headerFromStringInternal lines;
					super#seturl (self#lookup "get");
					super#process;

				method has lstr = Stringmap.mem lstr hmap

				method private status =
					try
						begin
							if self#has "http/1.1"
							then
								(List.nth (Str.split (regexp "[\t ]+") (Stringmap.find "http/1.1" hmap)) 0)
							else
								(List.nth (Str.split (regexp "[\t ]+") (Stringmap.find "http/1.0" hmap)) 0)
						end
					with x ->
						Logger.showex "connection.ml" 100 x "exception in status";
						"";

				method get = (List.nth (Str.split (Str.regexp "[\t ]+") (Stringmap.find "get" hmap)) 0)

				method hasGet = self#has "get"
				method private hasHead = self#has "head"
				method private hasPost = self#has "post"
				method private hasPut = self#has "put"
				method private hasDelete = self#has "delete"
				method private hasTrace = self#has "trace"
				method private hasContentLength = self#has "content-length:"
				method private hasTransferEncoding = self#has "transfer-encoding:"
				method private hasContentType = self#has "content-type:"

				method contentLength =
					int_of_string (Stringmap.find "content-length:" hmap)

				method private transferEncoding =
					Stringmap.find "transfer-encoding:" hmap

				method connectionClose =
					Stringmap.mem "connection:" hmap && (Stringmap.find "connection:" hmap) <> "close"

				method virtual transferType : tPacketType

				method host =
					try
						let h = self#lookup "host:" in
							Logger.show 3 "connection.ml" 129 ("result of lookup of host="^h);
							h
					with x ->
						Logger.show 3 "connection.ml" 132 "lookup failed, returning blank string";
						""

				method port =
					try
						(int_of_string (self#lookup "port"))
					with x ->
						80

				method transferType =
					if transactionType == Request
					then
						HeaderOnly
					else
						let lstatus = self#status in
							if (Str.string_match (regexp "204|304") lstatus 0 == true) || ((int_of_string lstatus) < 200)
							then
								HeaderOnly
							else if (self#hasContentLength && self#contentLength = 0)
							then
								HeaderOnly
							else if (self#hasTransferEncoding && self#transferEncoding = "chunked")
							then
								Chunked
							else if (self#hasContentLength)
							then
								ContentLength
							else if (self#hasContentType)
							then
								ContentType
							else
								Eof;
			end;;

		let reReturn = (Str.regexp "\r");;

		class connection (clientfd:Mis.fdAndChannel) (cinfo:Clientinfo.clientInfo) =
		object(self)
			val mutable serverfd = new Mis.fdAndChannel;
			val hrequest = new httpDataTransfer Request;
			val hresponse = new httpDataTransfer Response;
			val mutable packet = String.create 2048

			method connectionClose = hresponse#connectionClose

			method host = hrequest#host
			method getRequestHeader = hrequest

			method closeIfConnectionClose =
				if self#connectionClose then self#close

			method close =
				Unix.close clientfd#asfd;
				if serverfd#isopen then
					Unix.close serverfd#asfd;

			method setServerfd f = serverfd <- f

			method closed = (serverfd#isopen = false)

			method logdata astr =
				match cinfo#lchan with
						Mis.NilChannel -> ignore()
					| Mis.OutChannel oc -> Printf.fprintf oc "%s\n" astr;

			method doSelect =
				let (rd,wr,ex) = Unix.select [clientfd#asfd;serverfd#asfd;] [] [] 0.0 in
					(List.length rd) <> 0

			method readLine =
				let lstr = input_line (clientfd#asInchan) in
					lstr

			method readUnlimitedBlock =
				let lstr = String.create 2048 in
				let nchars = Unix.recv (clientfd#asfd) ~buf:lstr ~pos:0 ~len:(String.length lstr) ~mode:[] in
					String.copy (String.sub lstr 0 nchars);

			method readLimitedBlock maxsize =
				let readsize = (max maxsize cinfo#blocksz) in
				let lstr = String.create readsize in
				let nchars = Unix.recv (clientfd#asfd) ~buf:lstr ~pos:0 ~len:readsize ~mode:[] in
					String.copy (String.sub lstr 0 nchars);

			method readHeaderGen hdr (rab:Mis.readBufferToLines) =
				hdr#header_from_string (rab#lines());
				if hdr#hasGet then hdr#seturl hdr#get;

			method readHeader direction =
				Logger.show 3 "connection.ml" 221 ("readHeader");
				try
					let rab = new Mis.readBufferToLines clientfd#asInchan in
						match direction with
								Mis.ClientToServer -> self#readHeaderGen hrequest rab
							| Mis.ServerToClient -> self#readHeaderGen hresponse rab;
				with x ->
					Logger.showex "connection.ml" 228 x "exception in readHeader";
					raise x;

			method readBlockContentLength =
				let clength = hresponse#contentLength in
				let remainingLength = ref clength in
					while !remainingLength > 0 do
						let lstr = self#readLimitedBlock !remainingLength in
							remainingLength := (!remainingLength - (String.length lstr))
					done

			method readOneChunk =
				let lstr = self#readLine in
					if String.length lstr = 0 then false
					else let chunklen = int_of_string ("0x" ^ lstr) in
					let lstr2 = self#readLimitedBlock chunklen in
						packet <- (packet ^ lstr2);
						true

			method readBlockChunked =
				while self#readOneChunk do
					ignore();
				done

			method readBlockEof =
				while self#doSelect do
					self#readUnlimitedBlock
				done;
				self#close

			method readBlock =
				match hresponse#transferType with
						HeaderOnly -> ignore()  (* header already read  *)
					| Chunked -> self#readBlockChunked
					| ContentLength -> self#readBlockContentLength
					| _ -> self#readBlockEof


			method writeBlock =
				output_string serverfd#asOutchan packet;
				self#logdata packet;

			method writeToSource lstr =
				output_string clientfd#asOutchan lstr;
				self#logdata lstr;

			method readFromSource =
				if self#doSelect = false
				then
					raise NoData
				else
					input_line clientfd#asInchan

			method transferBlock direction =
				Logger.show 3 "connection.ml" 282 ("transferBlock");
				try
					self#readHeader direction;
					self#readBlock;
					self#writeBlock;
				with x -> Logger.showex "connection.ml" 296 x "exception caught in transfer block";

			method readAndDiscardHeader =
				let lstr = read_line() in
					match lstr with
							"" -> ignore()
						| lstring -> self#readAndDiscardHeader
		end;;

		class ['htype] httpTransfer conn hdr maxlen sfd dfd =
		object
			val header = (hdr : 'htype);
			val mutable databuffer = "";
			val sourcefd = sfd;
			val destfd = dfd;

			method transfer =
				let databuffer = conn#readHeader header in
					ignore();
		end;;

	end;;

