module Unix = UnixLabels
open UnixLabels

module Mis =
struct

	type tLogchan = NilChannel | OutChannel of out_channel;;
	type tDirection = ClientToServer | ServerToClient;;

	(* exceptions *)
	exception InvalidHandle;;
	exception InternalError;;

	let standardOutChannel = (OutChannel(Pervasives.stdout));;

	(* threads *)
  (*mutex/condition pair*)
  class mucv =
  object
    val mu = Mutex.create();
    val cv = Condition.create();
    method lock = Mutex.lock mu;
    method unlock = Mutex.unlock mu;
    method wait = Condition.wait cv mu;
    method signal = Condition.signal cv;
  end ;;

	(* network *)
	(* argument processing *)
  let argi name dest desc = (name, Arg.Int(fun x -> dest := x), desc);;
  let argsd name dest desc = (name, Arg.String(fun x -> dest := x), desc);;

	let arg argnum = (Array.get Sys.argv argnum);;

	(* file and socket handling *)
	let puts s = print_string s;print_newline();;
	let fputs fd s = output_string fd s;;

	type tFileDescr = FdNil | FileDescr of UnixLabels.file_descr;;
	type tInchan = InchanNil | Inchan of in_channel;;
	type tOutchan = OutchanNil | Outchan of out_channel;;

	class fdAndChannel =
	object(self)
		val mutable fd = FdNil
		val mutable inchan = InchanNil
		val mutable outchan = OutchanNil
		val mutable openflag = false;

		method isopen = openflag;

		method set afd =
			match afd with
					FdNil -> fd <- FdNil;
						inchan <- InchanNil;
						outchan <- OutchanNil;
						openflag <- false
				| FileDescr lfd -> fd <- (FileDescr lfd);
						inchan <- (Inchan (Unix.in_channel_of_descr lfd));
						outchan <- (Outchan (Unix.out_channel_of_descr lfd));
						openflag <- true

		method setfromoutchan (ch:out_channel) =
			inchan <- InchanNil;
			outchan <- (Outchan ch);
			fd <- (FileDescr (Unix.descr_of_out_channel ch));
			openflag = true;

		method setfromfd (fd:Unix.file_descr) =
			self#set (FileDescr fd);

		method asInchan =
			match inchan with
					InchanNil -> raise InvalidHandle
				| Inchan lin -> lin

		method asOutchan =
			match outchan with
					OutchanNil -> raise InvalidHandle
				| Outchan lout -> lout

		method asfd =
			match fd with
					FdNil -> raise InvalidHandle
				| FileDescr ltfd -> ltfd

		method id =
			let saddr = Unix.getpeername self#asfd in
				match saddr with
						ADDR_UNIX lstr -> lstr
					| ADDR_INET (iip, iport) -> Unix.string_of_inet_addr iip
	end

	let fdAndChannelFromFd afd =
		let fdc = new fdAndChannel in
			fdc#set afd;
			fdc;;

  class tcpSocket host port =
  object(self)
		inherit fdAndChannel as super
		val sock = Unix.socket ~domain:Unix.PF_INET ~kind:Unix.SOCK_STREAM ~protocol:0;
		val inaddr = Unix.ADDR_INET((Array.get ((Unix.gethostbyname host).h_addr_list) 0),port)
		val hentry = Unix.gethostbyname host;

		initializer
			Unix.connect sock inaddr;
			super#setfromfd sock;
	end

  (* Strings *)

	type maybeString = StringNil | MisString of string;;
  exception NilString

	class misString =
		object
			val mutable str = StringNil

			method set lstr =
				match lstr with
						StringNil -> raise NilString
					| MisString s -> str <- (MisString s);

			method setstr lstr = str <- (MisString lstr);

			method asString =
				match str with
						StringNil -> ""
					| MisString lstr -> lstr;

			method asInteger =
				match str with
						StringNil -> raise NilString
					| MisString lstr -> int_of_string lstr;
		end

	let rec stringOfStringList slist str =
		match slist with
				head :: rest -> stringOfStringList rest (str ^ head)
			| [] -> str;;

	let whitespace () = Str.regexp "[ \t\r\n]";;
	let spacetab () = Str.regexp "[ \t]";;

	class bufferToLines =
	object(self)
		val mutable buf = String.create 0;
		val mutable bufLessReturns = String.create 0;
		val mutable lines = [(String.create 0)];
		method lines () = lines;
		method set buf =
			bufLessReturns <- Str.global_replace (Str.regexp "\013") "" buf;
			lines <- Str.split (Str.regexp "\n") bufLessReturns;

		method getline () =
			match lines with
					first :: rest ->
						lines <- rest;
						first
				| [] -> "";
	end;;

	let readBufferSize () = 32256;;

	class readBufferToLines chan =
	object(self)
		val bufferToLines = new bufferToLines;
		val mutable buf = String.create 0;

		method lines () = bufferToLines#lines();

		method getline () = bufferToLines#getline();

		initializer
		let tbuf = String.create (readBufferSize()) in
		let bufsize = input chan tbuf 0 (readBufferSize()) in
			buf <- String.sub tbuf 0 bufsize;
			bufferToLines#set buf;
	end;;

end;;
